from django.http import HttpResponse, HttpResponseRedirect, Http404, \
    HttpResponseBadRequest
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django import forms
from django.db import transaction
from django.db.models import F, Min
from django.forms.models import modelform_factory, modelformset_factory
from django.views import generic
import itertools
import os

from .models import Program, Form, FormBlock, CustomBlock, CollectionBlock, \
    SubmissionRecord, SubmissionItem
from .forms import OpenForm, SubmissionForm, ItemFileForm, ItemsFormSet, \
    ItemsForm
from .filetype import FileType
from .signals import submission_handle_submit
from .utils import delete_file, get_file_extension, get_tooltips


class ProgramIndexView(generic.ListView):
    template_name = 'formative/index.html'
    context_object_name = 'programs'
    
    def get_queryset(self):
        return Program.objects.filter(hidden=False)


class ProgramView(generic.DetailView):
    model = Program
    template_name = 'formative/program.html'
    slug_field = 'slug'


class ProgramFormMixin(generic.edit.FormMixin):
    def dispatch(self, request, *args, **kwargs):
        form = get_object_or_404(Form.objects.select_related('program'),
                                 program__slug=self.kwargs['program_slug'],
                                 slug=self.kwargs['form_slug'])
        self.program_form = form
        
        if self.program_form.hidden(): # and 'sid' not in kwargs:
            key, params = self.program_form.access_enable(), self.request.GET
            template = self.template_name
            if 'thanks' not in template and 'continue' not in template:
                if not key or 'access' not in params or params['access'] != key:
                    kwargs = {'slug': self.program_form.program.slug}
                    return HttpResponseRedirect(reverse('program',
                                                        kwargs=kwargs))
        
        return super().dispatch(request, *args, **kwargs)
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        
        if not self.program_form.model: raise Http404()
        
        context['program_form'] = self.program_form
        context['tooltips'] = get_tooltips
        return context


class ProgramFormView(ProgramFormMixin, generic.edit.ProcessFormView,
                      generic.base.TemplateResponseMixin):
    template_name = 'formative/form.html'
    form_class = OpenForm
    context_object_name = 'program_form'
    
    def get_success_url(self):
        return reverse('form_continue', kwargs={
            'program_slug': self.program_form.program.slug,
            'form_slug': self.program_form.slug,
        })
    
    def form_valid(self, form):
        model = self.program_form.model
        self.object, created = model.objects.get_or_create(
            _email=form.cleaned_data['email']
        )
        
        if not self.object._submitted: template = 'continue'
        else: template = 'confirmation'
        
        self.object._send_email(form=self.program_form, name=template)
        return super().form_valid(form)


class SubmissionView(ProgramFormMixin, generic.UpdateView):
    template_name = 'formative/submission.html'
    context_object_name = 'submission'
    first_page = False
    
    def dispatch(self, request, *args, **kwargs):
        self.page = None
        if 'page' in self.kwargs: self.page = self.kwargs['page']
        if self.first_page: self.page = 1

        self.skipped, self.blocks_by_name, self.formsets = {}, {}, None
        
        return super().dispatch(request, *args, **kwargs)
    
    def get_object(self):
        submission = get_object_or_404(self.program_form.model,
                                       _id=self.kwargs['sid'])
        self.program_form.item_model
        
        return submission

    def get_form(self):
        fields, widgets, customs, stocks = [], {}, {}, {}

        self.query = self.program_form.visible_blocks()
        if self.page: self.query = self.query.filter(page=self.page)
        # having the form for the final GET simplifies templates for review step
        elif self.request.method == 'POST': self.query = self.query.none()

        blocks_checked, enabled = {}, []
        skipped_pages = self.object._skipped[:self.page or self.object._valid]
        skipped_ids = dict.fromkeys(itertools.chain(*skipped_pages), True)
        
        for block in self.query:
            d_id = block.dependence_id
            if d_id and d_id not in blocks_checked:
                # when we encounter a new block that some field is dependent on,
                # if it isn't among the blocks that were already _skipped,
                # enable this page's fields with a dependency matching the value
                if d_id not in skipped_ids:
                    b = FormBlock.objects.get(id=d_id)
                    if b.block_type() == 'stock':
                        stock = b.stock
                        values = { n: getattr(self.object, stock.field_name(n))
                                   for n in stock.widget_names() }
                        v = b.stock.conditional_value(**values)
                    elif b.block_type() == 'custom':
                        v = b.conditional_value(getattr(self.object, b.name))
                    else: # collection
                        v = bool(self.object._items.filter(_block=b.pk))
                    
                    enabled += b.enabled_blocks(v, self.page)
                blocks_checked[d_id] = True
            
            if d_id and block.id not in enabled:
                self.skipped[block.id] = block
                continue
            
            for name, field in block.fields():
                fields.append(name)
                if block.block_type != 'collection':
                    self.blocks_by_name[name] = block
                
                if block.block_type() == 'custom':
                    customs[name] = block
                    if block.type == CustomBlock.InputType.CHOICE:
                        widgets[name] = forms.RadioSelect
                    elif block.type == CustomBlock.InputType.BOOLEAN:
                        widgets[name] = forms.CheckboxInput
                elif block.block_type() == 'stock':
                    stocks[name] = block.stock
                    for name in block.stock.widget_names():
                        field_name = block.stock.field_name(name)
                        if (widget := block.stock.form_widget(name)):
                            widgets[field_name] = widget
        
        # this reuses self.query:
        self.formsets = self.get_formsets(enabled)
        
        def callback(model_field, **kwargs):
            name = model_field.name
            if name in customs:
                return customs[name].form_field(model_field, **kwargs)
            return model_field.formfield(**kwargs)
        
        form_class = modelform_factory(self.program_form.model,
                                       form=SubmissionForm,
                                       formfield_callback=callback,
                                       fields=fields, widgets=widgets)
        
        f = form_class(custom_blocks=customs, stock_blocks=stocks,
                       program_form=self.program_form, page=self.page,
                       **self.get_form_kwargs())
        return f
    
    def get_formsets(self, enabled):
        formsets = {}
        for block in self.query:
            if block.block_type() != 'collection': continue
            if block.dependence and block.id not in enabled: continue
            
            kwargs = self.get_form_kwargs()
            kwargs.pop('prefix')
            kwargs.pop('instance')

            item_model = self.program_form.item_model
            queryset = self.object._items.filter(_block=block.pk)
            queryset = queryset.exclude(_file='', _filesize__gt=0)
            
            extra = 0
            if block.fixed and not queryset:
                extra = block.num_choices()
                choices = [ {block.name1: c} for c in block.fixed_choices() ]
                kwargs['initial'] = choices
            
            fields = block.collection_fields()
            if block.fixed and queryset:
                if self.request.method == 'POST': fields = fields[1:]
            
            FormSet = modelformset_factory(item_model,
                                           formset=ItemsFormSet, form=ItemsForm,
                                           fields=fields,
                                           # TODO: use edit_only once available
                                           max_num=extra, can_delete=False,
                                           extra=extra, validate_max=False)

            formset = FormSet(prefix=f'items{block.pk}', queryset=queryset,
                              block=block, instance=self.object, **kwargs)
            formsets[block.pk] = formset
        
        return formsets
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        form = context['program_form']
        
        context['field_labels'] = form.field_labels()
        args, items = {'page': self.page, 'skip': self.skipped.keys()}, {}
        for item in form.visible_items(self.object, **args):
            if item._block not in items: items[item._block] = []
            items[item._block].append(item)
        context.update(visible_items=items, formsets=self.formsets)
        
        if self.page:
            context.update({
                'page': self.page,
                'prev_page': self.page > 1 and self.page - 1 or None,
                'visible_blocks': form.visible_blocks(**args)
            })
        else: context['prev_page'] = form.num_pages()
        
        return context
    
    def reset_skipped(self, ids=None):
        # when resetting the results of later pages that have been invalidated,
        # we haven't yet encountered these blocks, so we have to look them up:
        if ids: skipped = self.program_form.blocks.filter(id__in=ids)
        else: skipped = self.skipped.values()
        
        rec = None
        if self.program_form.item_model:
            items = self.object._items.filter(_block__in=list(self.skipped))
            if items.exclude(_file='').exists():
                rec = SubmissionRecord.objects.get(
                    submission=self.object.pk,
                    type=SubmissionRecord.RecordType.FILES
                )
        for block in skipped:
            if block.block_type() == 'collection':
                items = self.object._items.filter(_block=block.pk)
                for item in items:
                    if item._file:
                        delete_file(item._file)
                        rec.number = F('number') - item._filesize
                        rec.save()
                        rec.refresh_from_db() # clear the decrementer
                items.delete() # bulk is ok for ranked, because it's all of them
            else:
                val = None
                if block.block_type() == 'custom': val = block.default_value()
                for name, f in block.fields():
                    setattr(self.object, name, val)
    
    def new_page_valid(self, form):
        changed = [ self.blocks_by_name[n].id for n in form.changed_data ]
        if not changed: return self.object._valid # don't update _valid
        
        query = FormBlock.objects.filter(id__in=changed)
        query = query.annotate(pagemin=Min('dependent__page'))
        query = query.aggregate(min_pagemin=Min('pagemin'))
        
        if not query['min_pagemin']: return self.object._valid
        return query['min_pagemin'] - 1
    
    def form_valid(self, form):
        if not self.page:
            res = submission_handle_submit.send(self.program_form,
                                                submission=self.object)
            if res and res[0][1]: return res[0][1]
            
            # the draft submission will now be marked as submitted
            self.program_form.submit_submission(self.object)
            self.object._send_email(form=self.program_form, name='confirmation')
            
            return HttpResponseRedirect(reverse('form_thanks',
                                                kwargs=self.url_args(id=False)))
        
        self.object = form.save(commit=False)
        
        if self.page <= self.object._valid:
            self.object._valid = self.new_page_valid(form)
        else: self.object._valid = self.page
        
        if self.object._valid < len(self.object._skipped):
            self.object._skipped = self.object._skipped[:self.object._valid]
        
        if len(self.object._skipped) < self.page:
            # assumes only by 1; TODO: skipping an entire page or multiple pages
            self.object._skipped.append([])
        
        can_ignore = [ k for k, v in self.skipped.items()
                       if v.block_type() != 'custom' or not v.default_value() ]
        self.object._skipped[self.page-1] = can_ignore
        self.reset_skipped()
        
        for formset in self.formsets.values():
            block = formset.block
            
            if block.fixed and not formset.get_queryset():
                forms, choices = {}, block.fixed_choices()
                for form in formset.forms:
                    rank_key = form.add_prefix('_rank')
                    if rank_key in form.data:
                        forms[int(form.data[rank_key])] = form
                
                # create the items in the given order
                for i in range(block.num_choices()):
                    if i in forms:
                        if form.cleaned_data[block.name1] not in choices:
                            continue # form tampering
                        item = forms[i].save(commit=False)
                        item._block = block.pk
                        item._collection = block.name
                        item._submission = self.object
                        item.save()
            
            else: formset.save()
            
            items = self.object._items.filter(_block=formset.block.pk)
            failed_uploads = items.filter(_file='', _filesize__gt=0)
            # must do one at a time, because of the ranked model:
            for item in failed_uploads: item.delete()
            
            types = {}
            for item in items:
                if 'type' not in item._filemeta: continue
                filetype = FileType.by_type(item._filemeta['type'])()
                if filetype not in types: types[filetype] = []
                types[filetype].append(item)
                
            for filetype, items in types.items(): filetype.submitted(items)
        
        self.object.save()
        return HttpResponseRedirect(self.get_success_url())
    
    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        
        form = self.get_form()
        formsets = self.formsets
        if form.is_valid() and all(f.is_valid() for f in formsets.values()):
            return self.form_valid(form)
        else:
            return self.form_invalid(form)
    
    def url_args(self, id=True):
        args = {
            'program_slug': self.program_form.program.slug,
            'form_slug': self.program_form.slug
        }
        if id: args['sid'] = self.object._id
        return args
        
    def render_to_response(self, context):
        form = self.program_form
        
        if self.object._submitted:
            if self.page or not form.review_after_submit():
                url = reverse('form_thanks', kwargs=self.url_args(id=False))
                return HttpResponseRedirect(url)
            else: context['submitted'] = True
        
        if self.page and form.status != Form.Status.ENABLED:
            if not form.extra_time():
                if self.object._valid == form.num_pages():
                    url = reverse('submission_review', kwargs=self.url_args())
                else: url = reverse('program',
                                    kwargs={'slug': form.program.slug})
                return HttpResponseRedirect(url)
        
        if (
            self.page and context['page'] <= self.object._valid + 1
            or self.object._valid == form.num_pages() or self.object._submitted
        ):
            return super().render_to_response(context)
        
        # tried to skip ahead - go back to the last page that can be displayed
        name, args = 'submission_page', self.url_args()
        if self.object._valid: args['page'] = self.object._valid + 1
        else: name = 'submission'
        
        return HttpResponseRedirect(reverse(name, kwargs=args))
    
    def get_success_url(self):
        kwargs = self.url_args() 

        name = 'submission_page'
        if 'continue' in self.request.POST:
            if self.page == self.program_form.num_pages(): # we're done
                name = 'submission_review'
            else: kwargs['page'] = self.page + 1
        else:
            if self.page == 1: name = 'submission'
            else: kwargs['page'] = self.page
        
        return reverse(name, kwargs=kwargs)


class SubmissionBase(generic.View):
    def get_context_data(self, **kwargs):
        context = kwargs
        
        context['form_block'] = self.block
        return context
    
    def dispatch(self, request, *args, **kwargs):
        form = get_object_or_404(Form,
                                 program__slug=self.kwargs['program_slug'],
                                 slug=self.kwargs['form_slug'])
        self.program_form = form
        if not self.program_form.item_model: raise Http404()
        
        if form.status == Form.Status.DRAFT: return HttpResponseBadRequest()
        
        self.submission = get_object_or_404(self.program_form.model,
                                            _id=self.kwargs['sid'])
        if self.submission._submitted: return HttpResponseBadRequest()
        
        return super().dispatch(request, *args, **kwargs)


class SubmissionItemCreateView(SubmissionBase,
                               generic.base.TemplateResponseMixin):
    template_name = 'formative/collection_items_new.html'
    http_method_names = ['post']
    
    def get_form(self, **kwargs):
        if not self.block.has_file: return forms.Form(data={})
        return ItemFileForm(block=self.block, data=kwargs)
    
    def get_formset(self, ids):
        FormSet = modelformset_factory(self.program_form.item_model,
                                       formset=ItemsFormSet, form=ItemsForm,
                                       fields=self.block.collection_fields(),
                                       max_num=0, can_delete=False)
        
        queryset = self.submission._items.filter(_block=self.block.pk)
        queryset = queryset.filter(_id__in=ids)
        
        formset = FormSet(prefix=f'items{self.block.pk}', queryset=queryset,
                          block=self.block, instance=self.submission)
        return formset
    
    def post(self, request, *args, **kwargs):
        if 'block_id' not in self.request.POST:
            return HttpResponseBadRequest()
        
        self.block = get_object_or_404(CollectionBlock,
                                       form=self.program_form,
                                       pk=self.request.POST['block_id'])
        
        qs = self.submission._items.filter(_block=self.block.pk)
        nitems = qs.exclude(_file='', _filesize__gt=0).count()
        
        files, uploading = [], True
        for key, val in self.request.POST.items():
            if not key.startswith('filesize'): continue
            sizeval = key[len('filesize'):]
            size = None
            if sizeval.isdigit(): size = int(sizeval)
            if size is None: return HttpResponseBadRequest()
            files.append((val, size))

        if self.block.max_items and nitems >= self.block.max_items:
            if 'item_id' not in self.request.POST: return HttpResponse('')
        
        if 'item_id' in self.request.POST and len(files) != 1:
            return HttpResponseBadRequest()
        if not files:
            files.append((None, None))
            uploading = False
        
        items, ids = [], []
        for name, size in files:
            form = self.get_form(name=name, size=size)
            item = self.program_form.item_model(_submission=self.submission,
                                                _collection=self.block.name,
                                                _block=self.block.pk)
            if 'item_id' in self.request.POST:
                item = get_object_or_404(self.program_form.item_model,
                                         _submission=self.submission.pk,
                                         _id=self.request.POST['item_id'])
            else:
                nitems += 1
            
            if self.block.max_items and nitems > self.block.max_items: break
            if uploading and self.program_form.status == Form.Status.COMPLETED:
                if 'item_id' in self.request.POST:
                    ids.append(item._id)
                    items.append(item)
                break
            
            if not form.is_valid():
                item._error = True
                if form.has_error('name'):
                    msg = form.errors['name'][0]
                elif form.has_error('size'):
                    msg = form.errors['size'][0]
                else: msg = form.non_field_errors()[0]
                item._message = msg[:SubmissionItem._message_maxlen()]
            else:
                if item._file:
                    rec = SubmissionRecord.objects.get(
                        submission=self.submission.pk,
                        type=SubmissionRecord.RecordType.FILES
                    )
                    rec.number = F('number') - item._filesize
                    rec.save()
                    
                    item._file, item._filesize = '', 0
                    delete_file(item._file)
                
                if name:
                    if self.block.autoinit_filename():
                        setattr(item, self.block.name1, name[:name.rindex('.')])
                    item._file, item._filesize = '', size
                
                item._error, item._message = False, ''
            
            item.save()
            ids.append(item._id)
            items.append(item)
        
        c = self.get_context_data(items=items, uploading=uploading,
                                  formset=self.get_formset(ids),
                                  field_labels=self.program_form.field_labels())
        return self.render_to_response(c)


class SubmissionItemBase(SubmissionBase):
    def get_item(self, for_update=False):
        if 'item_id' not in self.request.POST: return HttpResponseBadRequest()
        
        if for_update:
            queryset = self.program_form.item_model.objects.select_for_update()
        else: queryset = self.program_form.item_model.objects.all()
        
        try:
            return queryset.get(_submission=self.submission.pk,
                                _id=self.request.POST['item_id'])
        except queryset.model.DoesNotExist:
            name = queryset.model._meta.object_name
            raise Http404(f'No {name} matches the given query.')


class SubmissionItemUploadView(SubmissionItemBase):
    http_method_names = ['post']
    
    def post(self, request, *args, **kwargs):
        item = self.get_item()
        
        block = get_object_or_404(CollectionBlock,
                                  form=self.program_form,
                                  pk=item._block)
        
        if 'file' not in self.request.FILES: return HttpResponseBadRequest()
        if item._error: return HttpResponseBadRequest()
        
        item._file = self.request.FILES['file']
        if item._file.size != item._filesize: return HttpResponseBadRequest()
        
        types = block.allowed_filetypes()
        
        item.save()
        path = item._file.path
        
        id_filename = os.path.join(os.path.dirname(path), 'id.txt')
        if not os.path.isfile(id_filename):
            with open(id_filename, 'w') as id_file:
                id_file.write(self.submission._email + '\n')
        
        filetype_class = FileType.by_extension(get_file_extension(path))
        # the extension was supposed to be already validated
        if not filetype_class and types: return HttpResponseBadRequest()
        
        msg_maxlen = SubmissionItem._message_maxlen()
        def item_error(item, msg):
            item._error, item._message = True, msg[:msg_maxlen]
            delete_file(item._file)
            item._file, item._filesize = '', 0
            return msg
        
        msg = ''
        if filetype_class:
            filetype = filetype_class()
            meta = filetype.meta(path)
            if 'error' in meta: msg = item_error(item, meta['error'])
            else:
                file_limits = block.file_limits()
                if filetype.TYPE in file_limits:
                    msg = filetype.limit_error(meta, file_limits[filetype.TYPE])
                    if msg: item_error(item, msg)
                    else: msg = ''
                
                if not item._error:
                    opts = block.process_options(filetype.TYPE)
                    newmeta = filetype.process(item._file, meta, **opts)
                    if 'error' in newmeta:
                        msg = item_error(item, newmeta['error'])
                    else:
                        if 'message' in newmeta:
                            warn_msg = newmeta.pop('message')
                            item._message = warn_msg[:msg_maxlen]
                        if 'update_filesize' in newmeta:
                            item._filesize = newmeta.pop('update_filesize')
                        item._filemeta = newmeta
            item.save()
            
        rec, created = SubmissionRecord.objects.get_or_create(
            program=self.program_form.program, form=self.program_form.slug,
            submission=self.submission.pk,
            type=SubmissionRecord.RecordType.FILES
        )
        if created: rec.number = item._filesize
        else: rec.number = F('number') + item._filesize
        rec.save()
        
        return HttpResponse(msg)


class SubmissionItemRemoveView(SubmissionItemBase):
    http_method_names = ['post']
    
    def post(self, request, *args, **kwargs):
        file = None
        with transaction.atomic():
            item = self.get_item(for_update=True)
            file, filesize = item._file, item._filesize
            item.delete()
        
        if file:
            rec = SubmissionRecord.objects.get(
                submission=self.submission.pk,
                type=SubmissionRecord.RecordType.FILES
            )
            rec.number = F('number') - filesize
            rec.save()
            
            delete_file(file)
        
        return HttpResponse('')


class SubmissionItemMoveView(SubmissionItemBase):
    http_method_names = ['post']
    
    def post(self, request, *args, **kwargs):
        if 'rank' not in self.request.POST: return HttpResponseBadRequest()
        rank = self.request.POST['rank']
        if not rank.isdigit() or int(rank) <= 0:
            return HttpResponseBadRequest()
        
        item = self.get_item()
        item._rank = int(rank)
        
        item.save()
        return HttpResponse('')



from django.utils.translation import gettext_lazy as _
import logging
from math import ceil, floor

__all__ = ["ImageFile", "DocumentFile", "AudioFile", "VideoFile"]


class FileType:
    types = {}
    extensions = {}
    composite = False
    logger = logging.getLogger('django.request')
    
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.TYPE:
            cls.types[cls.TYPE] = cls
            for extension in cls.EXTENSIONS:
                cls.extensions[extension] = cls
    
    @classmethod
    def by_type(cls, type):
        return cls.types[type]
    
    @classmethod
    def by_extension(cls, extension):
        if extension in cls.extensions:
            return cls.extensions[extension]
        return None
    
    # TODO: option to enable or disable extensions supported by the type
    
    def allowed_extensions(self): return self.EXTENSIONS
    
    def meta(self, path): return {'type': self.TYPE}
    
    def limit_error(self, meta, limits):
        for key, val in limits.items():
            if not (key.startswith('max_') or key.startswith('min_')): continue
            name = key[4:]
            if name == 'filesize' or name not in meta: continue
            v = meta[name]
            if type(v) == float: formatted_v = f'{v:.4f}'
            else: formatted_v = v
            
            retname = _(name.replace('_', ' '))
            if key.startswith('max_'):
                if v > val:
                    m = _('Maximum for %(name)s is %(maxval)s. It is %(val)s.')
                    return m % {'name': retname, 'maxval': floor(val),
                                'val': formatted_v}
            else:
                if v < val:
                    m = _('Minimum for %(name)s is %(minval)s. It is %(val)s.')
                    return m % {'name': retname, 'minval': ceil(val),
                                'val': formatted_v}
        
        return None
    
    def process(self, file, meta, **kwargs): return meta
    
    def submitted(self, items): pass
    
    def admin_limit_fields(self): return ()
    
    def admin_total_fields(self): return ()
    
    def admin_processing_fields(self): return ()
    
    def artifact_url(self, name, file_url): return None


from .image import ImageFile
from .document import DocumentFile
from .av import AudioFile, VideoFile



from django import forms
from django.core.exceptions import ValidationError
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.formats import date_format, time_format
from django.utils.translation import gettext_lazy as _
from datetime import timedelta

from ..models import Form, CustomBlock, SubmissionItem
from ..filetype import FileType
from ..validators import MinWordsValidator, MaxWordsValidator, \
    FileExtensionValidator, FileSizeValidator
from ..utils import get_file_extension, human_readable_filesize


class OpenForm(forms.Form):
    email = forms.EmailField(label='email address')


class SubmissionForm(forms.ModelForm):
    def __init__(self, program_form=None, custom_blocks=None, stock_blocks=None,
                 page=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.program_form = program_form
        self.custom_blocks = custom_blocks
        self.stock_blocks = stock_blocks
        self.page = page
        
        for name, field in self.fields.items():
            if name in stock_blocks:
                stock = stock_blocks[name]
                
                widget = stock.get_widget(name)
                field.validators += stock.field_validators(widget)
                
                if stock.field_required(widget): field.required = True
                
                if type(field) == forms.TypedChoiceField:
                    if not field.initial: field.choices = field.choices[1:]
                continue
            
            block = custom_blocks[name]
            if type(field) == forms.TypedChoiceField:
                # TODO: report bug with passing empty_label to .formfield()
                field.choices = field.choices[1:]
            elif type(field) == forms.CharField:
                if block.min_words:
                    field.validators.append(MinWordsValidator(block.min_words))
                if block.max_words:
                    field.validators.append(MaxWordsValidator(block.max_words))

    def clean(self):
        super().clean()
        cleaned_data = self.cleaned_data
        
        if not self.page and self.program_form.status != Form.Status.ENABLED:
            msg, params = _('Application cannot be submitted. '), {}
            if self.program_form.status == Form.Status.COMPLETED:
                if self.program_form.extra_time(): msg = None
                else: msg += _('This form closed on %(date)s at %(time)s.')
                
                comp = self.program_form.completed
                params = {
                    'time': time_format(comp.time(), format='TIME_FORMAT'),
                    'date': date_format(comp.date(), format='SHORT_DATE_FORMAT')
                }
            else: msg += _('Submissions are temporarily disabled.')
            
            if msg:
                self.add_error(None, ValidationError(msg, params=params))
                return
        
        stocks = {}
        for name, field in self.fields.items():
            if name in self.custom_blocks:
                block = self.custom_blocks[name]
                if self.has_error(name): continue
                
                try: data = block.clean_field(cleaned_data[name], field)
                except ValidationError as e: self.add_error(name, e.message)
                else: self.cleaned_data[name] = data
                
                if block.type in (CustomBlock.InputType.TEXT,
                                  CustomBlock.InputType.CHOICE):
                    # NULL for fields that're never seen; '' for no choice made
                    if cleaned_data[name] is None: self.cleaned_data[name] = ''
        
            elif name in self.stock_blocks:
                stock = self.stock_blocks[name]
                if stock.name not in stocks: stocks[stock.name] = [stock, {}]
                
                if self.has_error(name):
                    stocks[stock.name][0] = False
                    continue
                widget = stock.get_widget(name)
                if widget: stocks[stock.name][1][widget] = cleaned_data[name]
                else: stocks[stock.name][1] = cleaned_data[name]
            
        for name, (stock, data) in stocks.items():
            if not stock: continue # show validator errors on form fields first
            
            fields = stock.clean(data)
            
            if not stock.composite:
                if type(fields) == ValidationError:
                    self.add_error(name, err)
                else: self.cleaned_data[name] = fields
                continue
            err = False
            for widget, val in fields.items():
                if type(val) == ValidationError:
                    if widget: self.add_error(stock.field_name(widget), val)
                    else: self.add_error(None, val)
                    err = True
            if not err:
                for widget, val in fields.items():
                    self.cleaned_data[stock.field_name(widget)] = val


class ItemFileForm(forms.Form):
    name = forms.CharField(max_length=SubmissionItem._filename_maxlen(),
                           error_messages={
        'max_length': _('File name cannot exceed %(limit_value)d characters.')
    })
    size = forms.IntegerField()
    
    def __init__(self, block=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        self.block = block
        
        if block.file_optional:
            self.fields['name'].required = False
            self.fields['size'].required = False
        
        maxsize = block.max_filesize()
        if maxsize:
            self.fields['size'].validators.append(FileSizeValidator(maxsize))
        
        extensions = self.block.allowed_extensions()
        validator = FileExtensionValidator(allowed_extensions=extensions)
        self.fields['name'].validators.append(validator)
        
    def clean(self):
        super().clean()
        cleaned_data = self.cleaned_data
        if 'name' not in cleaned_data: return
        
        extension = get_file_extension(self.cleaned_data['name'])
        filetype = FileType.by_extension(extension)
        limits = self.block.file_limits()
        
        if filetype and filetype.TYPE in limits:
            if 'max_filesize' in limits[filetype.TYPE]:
                maxval = limits[filetype.TYPE]['max_filesize']
                if cleaned_data['size'] > maxval:
                    msg = _('Maximum file size for this type of file is '
                            '%(limit_value)s.')
                    params = {'limit_value': human_readable_filesize(maxval)}
                    raise ValidationError(msg, code='max_value', params=params)
            
            if 'min_filesize' in limits[filetype.TYPE]:
                maxval = limits[filetype.TYPE]['min_filesize']
                if cleaned_data['size'] < maxval:
                    msg = _('Minimum file size for this type of file is '
                            '%(limit_value)s.')
                    params = {'limit_value': human_readable_filesize(minval)}
                    raise ValidationError(msg, code='min_value', params=params)


class ItemsForm(forms.ModelForm):
    def __init__(self, block=None, field_blocks=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.block = block
        self.field_blocks = field_blocks
        
        for name, f in self.fields.items():
            if name in self.field_blocks: field_block = self.field_blocks[name]
            else: field_block = CustomBlock.text_create()
            
            if field_block.min_words:
                f.validators.append(MinWordsValidator(field_block.min_words))
            if field_block.max_words:
                f.validators.append(MaxWordsValidator(field_block.max_words))
    
    def clean(self):
        super().clean()
        
        if self.instance and self.instance._error:
            for name in self.fields.keys():
                if name in self.errors: self.errors.pop(name)
                self.fields[name].required = False # TODO move


class ItemsFormSet(forms.BaseModelFormSet):
    def __init__(self, block=None, instance=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.block, self.instance = block, instance
        
        cfields = block.collection_fields()
        blocks = block.form.custom_blocks().filter(page=0, name__in=cfields)
        self.field_blocks = { b.name: b for b in blocks }
    
    def get_form_kwargs(self, index):
        kwargs = super().get_form_kwargs(None) # it just makes a copy
        
        kwargs['block'] = self.block
        kwargs['field_blocks'] = self.field_blocks
        return kwargs
    
    # the way Django formsets index forms in POST data doesn't work well when
    # there's AJAX on the page that can do insertions and deletions.
    # we assume here that the forms in the formset always have a pk'd instance,
    # except for unbound formsets, which can have extra (for fixed collections)
    @cached_property
    def forms(self):
        args_list, objects = [], { str(o.pk): o for o in self.get_queryset() }
        creating = False
        if self.block.fixed and not objects:
            creating = True
            objects = [ str(i) for i in range(self.block.num_choices()) ]
        
        defaults = {
            'auto_id': self.auto_id,
            'error_class': self.error_class,
            'renderer': self.renderer
        }
        def new_kwargs():
            kwargs = self.get_form_kwargs(None)
            kwargs.update(defaults)
            return kwargs
        
        if self.is_bound:
            for key, val in self.data.items():
                if not key.startswith(self.prefix + '-'): continue
                
                if creating: pk_name = '_rank'
                else: pk_name = '_id'
                
                rest = key[len(self.prefix)+1:]
                if not rest.endswith('-' + pk_name): continue
                
                pk_str = rest[:-len(pk_name)-1]
                
                if pk_str in objects:
                    kwargs = new_kwargs()
                    
                    kwargs['prefix'] = self.prefix + '-' + pk_str
                    kwargs['data'], kwargs['files'] = self.data, self.files
                    
                    if not creating:
                        kwargs['instance'] = objects[pk_str]
                    args_list.append(kwargs)
        elif not creating:
            for instance in self.get_queryset():
                kwargs = new_kwargs()
                
                kwargs['instance'] = instance
                kwargs['prefix'] = self.prefix + '-%s' % (instance.pk,)
                args_list.append(kwargs)
        else:
            for i in range(self.extra):
                kwargs = new_kwargs()
                
                kwargs['prefix'] = self.prefix + '-%s' % (i,)
                if self.initial_extra: kwargs['initial'] = self.initial_extra[i]
                args_list.append(kwargs)
        
        forms = []
        for i, kwargs in enumerate(args_list):
                form = self.form(**kwargs)
                self.add_fields(form, i)
                
                if not creating:
                    form.fields[self.model._meta.pk.name].required = True
                forms.append(form)
        
        return forms
    
# still need that management form, for now
#    def initial_form_count(self):
#        if not self.is_bound: return len(self.get_queryset())
#        return len(self.forms)
#    
#    def total_form_count(self):
#        if not self.is_bound: return 0
#        else: return self.initial_form_count()
    
    def clean(self):
        super().clean()
        
        files = self.instance._items.filter(_block=self.block.pk)
        files = files.exclude(_file='', _filesize__gt=0)
        file_errors = files.values_list('_error', flat=True)
        n = len(file_errors)
        if True in file_errors:
            msg = _('Some files have errors.')
            raise ValidationError(msg, code='file_error')
        
        self.valid_num(n)
        
        file_limits = self.block.file_limits()
        if 'total' not in file_limits: return
        
        for key, limit_val in file_limits['total'].items():
            if not (key.startswith('max_') or key.startswith('min_')): continue
            name = key[4:]
            
            if name == 'filesize': name = 'size'
            generic = FileType()
            if name == 'size':
                total = sum(files.values_list('_filesize', flat=True))
            else:
                total = sum(f._filemeta[name] for f in files
                            if name in f._filemeta)
            
            meta = { name: total }
            err = generic.limit_error(meta, {key: limit_val})
            if err: raise ValidationError(err)
    
    def valid_num(self, n):
        if not self.block.min_items and not self.block.max_items: return
        
        params, msg = {}, ''
        if self.block.min_items and self.block.max_items is None:
            if n >= self.block.min_items: return
            
            if self.block.has_file and not self.block.file_optional:
                msg = _('Number of files must be at least %(min_val)d.')
            else: msg = _('Number of items must be at least %(min_val)d.')
            params['min_val'] = self.block.min_items
        elif self.block.max_items is not None and not self.block.min_items:
            if n <= self.block.max_items: return
            
            if self.block.has_file and not self.block.file_optional:
                msg = _('Number of files must be at most %(max_val)d.')
            else: msg = _('Number of items must be at most %(max_val)d.')
            params['max_val'] = self.block.max_items
        else:
            if n >= self.block.min_items and n <= self.block.max_items: return
            
            if self.block.has_file and not self.block.file_optional:
                msg = _('Number of files must be between '
                        '%(min_val)d and %(max_val)d.')
            else: msg = _('Number of items must be between '
                          '%(min_val)d and %(max_val)d.')
            params['min_val'] = self.block.min_items
            params['max_val'] = self.block.max_items
        
        raise ValidationError(msg, code='invalid_num_items', params=params)



from django.utils.translation import gettext_lazy as _
import ffmpeg
import os

from ..utils import get_file_extension, thumbnail_path, subtitle_path
from . import FileType


class AVFileType(FileType):
    TYPE = None
    
    def meta(self, path):
        ret = super().meta(path)
        
        try:
            probe = ffmpeg.probe(path)
            streams, msg = probe['streams'], ''
            audio_streams = [ s for s in streams if s['codec_type'] == 'audio' ]
            video_streams = [ s for s in streams if s['codec_type'] == 'video' ]
            subtitles = [ s for s in streams if s['codec_type'] == 'subtitle' ]
            ext = get_file_extension(path)
            
            format = probe['format']['format_name']
            if ext == 'mp3':
                if 'mp3' not in format.split(','):
                    return {'error': _('Incorrect format for MP3 '
                                       'file extension.')}
            else:
                if 'mp4' not in format.split(','):
                    return {'error': _('Media container format must be MP4.')}
            
            if len(audio_streams) > 1:
                return {'error': _('File contains more than 1 audio stream.')}
            if len(video_streams) > 1:
                return {'error': _('File contains more than 1 video stream.')}
            ret.update(seconds=float(probe['format']['duration']),
                       bit_rate=float(probe['format']['bit_rate']))
            
            if audio_streams: ret.update(self.audio_meta(audio_streams[0], ext))
            if 'error' in ret: return ret
            
            if video_streams: ret.update(self.video_meta(video_streams[0], ext))
            if 'error' in ret: return ret
            
            if subtitles and video_streams:
                ret.update(self.subtitle_meta(subtitles, ext))
                if 'error' in ret: return ret
            
            if not audio_streams and not video_streams: return self.no_stream()
            if ext in VideoFile.EXTENSIONS and not video_streams:
                return self.no_stream()
            if ext in AudioFile.EXTENSIONS and not audio_streams:
                return self.no_stream()
            
            return ret
        
        except:
            if self.TYPE == 'audio':
                msg = _('Error occurred reading the audio file.')
            else: msg = _('Error occurred reading the video file.')
            self.logger.critical(msg, exc_info=True)
            return {'error': msg}


class AudioFile(AVFileType):
    TYPE = 'audio'
    EXTENSIONS = ('mp3', 'm4a')
    
    def no_stream(self):
        return {'error': _('No audio stream present in the audio file.')}
    
    def video_meta(self, stream, extension):
        if 'attached_pic' in stream['disposition']:
            if stream['disposition']['attached_pic']: return {}
        
        return {'error': _('Video stream present in the audio file.')}
    
    def audio_meta(self, stream, extension):
        codec = stream['codec_name']
        msg = _('Audio codec must be %(req)s. It is %(codec)s')
        if extension == 'mp3' and codec != extension:
            return {'error': msg % {'req': 'mp3', 'codec': codec}}
        if extension == 'm4a' and codec != 'aac':
            return {'error': msg % {'req': 'aac', 'codec': codec}}
        
        return {
            'audio_codec': codec,
            'audio_bitrate': float(stream['bit_rate']),
            'audio_samplerate': float(stream['sample_rate']),
            'audio_channels': stream['channels']
        }
    
    def subtitle_meta(self, stream, extension):
        return {}
    
    def admin_limit_fields(self):
        return ('seconds', 'bitrate', 'audio_bitrate', 'audio_samplerate',
                'audio_channels')


class VideoFile(AudioFile):
    TYPE = 'video'
    EXTENSIONS = ('mp4', 'm4v')
    
    def no_stream(self):
        return {'error': _('No video stream present in the video file.')}

    def video_meta(self, stream, extension):
        codec = stream['codec_name']
        if codec != 'h264':
            msg = _('Video codec must be H.264 (aka AVC). It is %(codec)s.')
            return {'error': msg % {'codec': codec}}
        
        if 'frame_rate' in stream: frame_rate = stream['frame_rate']
        else: frame_rate = stream['avg_frame_rate']
        divisor = float(frame_rate.split('/')[1])
        if not divisor:
            return {'error': _('Video stream has an invalid frame rate.')}
        
        return {
            'video_codec': codec,
            'video_bitrate': float(stream['bit_rate']),
            'video_framerate': float(frame_rate.split('/')[0]) / divisor,
            'video_height': stream['height'],
            'video_width': stream['width']
        }
    
    def subtitle_meta(self, streams, extension):
        subs = []
        for i, stream in enumerate(streams):
            codec = stream['codec_name']
            if codec != 'mov_text':
                msg = _('Subtitle format must be mov_text (aka MP4TT). '
                        'It is %(codec)s.')
                return {'error': msg % {'codec': codec}}
        
            lang = 'und'
            if 'language' in stream['tags']: lang = stream['tags']['language']
            
            default = False
            if 'disposition' in stream and 'default' in stream['disposition']:
                default = bool(stream['disposition']['default'])
            
            subs.append((i, lang, default))
        
        return { 'subtitle_streams': subs }
    
    def process(self, file, meta, extract_captions=True, **kwargs):
        if not extract_captions or 'subtitle_streams' not in meta: return meta
        
        try:
            v, subs = ffmpeg.input(file.path), []
            for idx, lang, default in meta['subtitle_streams']:
                filename = subtitle_path(file.path, lang)
                v = v.output(filename, map=f'0:s:{idx}')
                relpath = subtitle_path(file.name, lang)
                
                desc = {'language': lang, 'file': relpath}
                if default: desc['default'] = True
                subs.append(desc)
            v.run()
            
            meta['subtitles'] = subs
            meta.pop('subtitle_streams')
            msg = _('Extracted %(num)d embedded subtitle track(s).')
            meta['message'] = msg % {'num': len(subs)}
            
            return meta
        except:
            msg = _('Error occurred processing the video file.')
            self.logger.critical(msg, exc_info=True)
            return {'error': msg}
    
    def submitted(self, items):
        for item in items:
            file = item._file
            
            try:
                v = ffmpeg.input(file.path, ss=item._filemeta['seconds']/4)
                outpath = thumbnail_path(file.path, ext='jpg')
                if not os.path.isfile(outpath):
                    v.filter('scale', 120, -1).output(outpath, vframes=1).run()
            except:
                self.logger.critical('Error generating video thumbnail.',
                                     exc_info=True)
    
    def admin_limit_fields(self):
        return ('seconds', 'bitrate', 'video_bitrate', 'video_framerate',
                'video_width', 'video_height')
    
    def admin_total_fields(self):
        return ('seconds',)
    
    def artifact_url(self, name, url):
        if not name.startswith('subtitles_'): return None
        return subtitle_path(url, name[len('subtitles_'):])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['formative',
 'formative.filetype',
 'formative.forms',
 'formative.migrations',
 'formative.models',
 'formative.stock',
 'formative.templatetags']

package_data = \
{'': ['*']}

install_requires = \
['django-better-admin-arrayfield>=1.4.2,<2.0.0',
 'django-environ',
 'django-jazzmin',
 'django-localflavor>=3.1,<4.0',
 'django-polymorphic',
 'django-webpack-loader',
 'django-widget-tweaks',
 'django>=4.0,<5.0',
 'ffmpeg-python',
 'gunicorn',
 'markdown',
 'markdown-link-attr-modifier>=0.2.0,<0.3.0',
 'pikepdf>=5.1.1',
 'pillow',
 'psycopg2',
 'whitenoise>=6.0.0,<7.0.0']

extras_require = \
{':python_version >= "3.8" and python_version < "3.9"': ['backports.zoneinfo'],
 'reviewpanel': ['reviewpanel>=0.5.3,<0.6.0']}

setup_kwargs = {
    'name': 'formative',
    'version': '0.9.0',
    'description': 'Self-hosted web app for collecting form responses and files',
    'long_description': '**Formative** is a self-hosted web app for collecting form responses and\nfiles. For application forms, you can use it together with\n[reviewpanel](https://github.com/johncronan/reviewpanel) to review and score\napplicant submissions.\n\nDevelopment of Formative was funded by\n[Public Media Institute](https://publicmediainstitute.com), with support from\n[The Andy Warhol Foundation for the Visual Arts](https://warholfoundation.org).\n',
    'author': 'John Kyle Cronan',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/johncronan/formative',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from django.urls import path

from . import views

urlpatterns = [
    path('', views.ProgramIndexView.as_view(), name='program_index'),
    path('<slug:slug>/', views.ProgramView.as_view(), name='program'),
    path('<slug:program_slug>/<slug:form_slug>/',
         views.ProgramFormView.as_view(), name='form'),
    path('<slug:program_slug>/<slug:form_slug>/continue',
         views.ProgramFormView.as_view(template_name='formative/continue.html'),
         name='form_continue'),
    path('<slug:program_slug>/<slug:form_slug>/thanks',
         views.ProgramFormView.as_view(template_name='formative/thanks.html'),
         name='form_thanks'),
    path('<slug:program_slug>/<slug:form_slug>/<uuid:sid>/',
         views.SubmissionView.as_view(first_page=True), name='submission'),
    path('<slug:program_slug>/<slug:form_slug>/<uuid:sid>/page-<int:page>',
         views.SubmissionView.as_view(), name='submission_page'),
    path('<slug:program_slug>/<slug:form_slug>/<uuid:sid>/review',
         views.SubmissionView.as_view(template_name='formative/review.html'),
         name='submission_review'),
    path('<slug:program_slug>/<slug:form_slug>/<uuid:sid>/item',
         views.SubmissionItemCreateView.as_view(), name='item'),
    path('<slug:program_slug>/<slug:form_slug>/<uuid:sid>/file',
         views.SubmissionItemUploadView.as_view(), name='item_file'),
    path('<slug:program_slug>/<slug:form_slug>/<uuid:sid>/removeitem',
         views.SubmissionItemRemoveView.as_view(), name='item_remove'),
    path('<slug:program_slug>/<slug:form_slug>/<uuid:sid>/moveitem',
         views.SubmissionItemMoveView.as_view(), name='item_move'),
]



from django.db import models
from django.forms import DateInput

from . import StockWidget

class DateWidget(StockWidget):
    TYPE = 'date'
    
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        
        self.template_name = 'date.html'
        self.review_template_name = 'review.html'
    
    def fields(self):
        field = models.DateField(null=True, blank=True)
        
        return [(self.field_name(), field)]
    
    def form_widget(self, name):
        return DateInput(format='%m/%d/%Y')



from django.db import models, connection
from django.db.models import Q, Max, Case, Value, When, Exists, OuterRef, \
    UniqueConstraint, Subquery
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import FieldError, ValidationError
from django.template import Template, loader
from django.utils.functional import cached_property
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from django.urls import reverse
from polymorphic.models import PolymorphicModel
import uuid
import markdown
from markdown_link_attr_modifier import LinkAttrModifierExtension
from itertools import groupby
from pathlib import Path
from datetime import timedelta
import os

from ..stock import StockWidget
from ..filetype import FileType
from ..utils import create_model, remove_p, send_email, submission_link, \
    thumbnail_path
from .ranked import RankedModel, UnderscoredRankedModel
from .automatic import AutoSlugModel


class Program(AutoSlugModel):
    class Meta:
        ordering = ['created']
    
    name = models.CharField(max_length=64)
    slug = models.SlugField(max_length=30, unique=True, allow_unicode=True,
                            verbose_name='identifier')
    db_slug = models.SlugField(max_length=30, unique=True, allow_unicode=True,
                               editable=False)
    description = models.CharField(max_length=250, blank=True)
    options = models.JSONField(default=dict, blank=True)
    hidden = models.BooleanField(default=False)
    created = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return self.name
    
    def validate_unique(self, exclude=None):
        super().validate_unique(exclude=exclude)
        if not self._state.adding: return
        
        if Program.objects.filter(db_slug=self.slug.replace('-', '')).exists():
            msg = 'Identifier (with hyphens removed) must be unique.'
            raise ValidationError(msg)
    
    @cached_property
    def markdown(self):
        return markdown.Markdown(extensions=[
            LinkAttrModifierExtension(new_tab='external_only')
        ])
    
    def visible_forms(self):
        pub = self.forms.exclude(status=Form.Status.DRAFT)
        return pub.exclude(options__hidden__isnull=False)
    
    def home_url(self):
        if 'home_url' in self.options: return self.options['home_url']
        return None


class Form(AutoSlugModel):
    class Meta:
        constraints = [
            UniqueConstraint(fields=['program', 'slug'], name='unique_slug'),
            UniqueConstraint(fields=['program', 'db_slug'],
                             name='unique_db_slug')
        ]
    
    class Status(models.TextChoices):
        DRAFT = 'draft', _('unpublished')
        DISABLED = 'disabled', _('submissions disabled')
        ENABLED = 'enabled', _('published/enabled')
        COMPLETED = 'completed', _('completed')
    
    class Validation(models.TextChoices):
        # currently, validation to create a submission is always by email
        EMAIL = 'email', _('email address')
    
    program = models.ForeignKey(Program, models.CASCADE,
                                related_name='forms', related_query_name='form')
    name = models.CharField(max_length=64)
    slug = models.SlugField(max_length=30, allow_unicode=True,
                            verbose_name='identifier')
    db_slug = models.SlugField(max_length=30, allow_unicode=True,
                               editable=False)
    status = models.CharField(max_length=16, default=Status.DRAFT,
                              choices=Status.choices)
    options = models.JSONField(default=dict, blank=True)
    created = models.DateTimeField(auto_now_add=True)
    # this is also the published date, for enabled and completed forms
    modified = models.DateTimeField(default=timezone.now, editable=False)
    completed = models.DateTimeField(null=True, blank=True, editable=False)
    validation_type = models.CharField(max_length=16, editable=False,
                                       default=Validation.EMAIL,
                                       choices=Validation.choices)
    
    def __str__(self):
        return self.name
    
    def validate_unique(self, exclude=None):
        super().validate_unique(exclude=exclude)
        if not self._state.adding: return
        
        if Form.objects.filter(program=self.program,
                               db_slug=self.slug.replace('-', '')).exists():
            msg = 'Identifier (with hyphens removed) must be unique ' \
                  'within this program.'
            raise ValidationError(msg)
    
    @cached_property
    def model(self):
        if self.status == self.Status.DRAFT: return None
        
        fields = []
        for block in self.blocks.exclude(page=0, _rank__gt=1):
            fields += block.fields()
        
        name = self.program.db_slug + '_' + self.db_slug
        class Meta:
            verbose_name = self.slug + ' submission'
            verbose_name_plural = self.slug + ' submissions'
        return create_model(name, fields, program=self.program.db_slug,
                            base_class=Submission, meta=Meta)
    
    @cached_property
    def item_model(self):
        if self.status == self.Status.DRAFT: return None
        
        collections = self.collections()
        if not collections: return None
        
        names = []
        for c in collections:
            for field in c.collection_fields():
                if field not in names: names.append(field)
        
        fields = [
            # the first column links submission items to the submission
            ('_submission', models.ForeignKey(self.model, models.CASCADE,
                                              related_name='_items',
                                              related_query_name='_item'))
        ]
        
        field_blocks = { b.name: b for b in
                         self.custom_blocks().filter(page=0, _rank__gt=1) }
        for n in names:
            # look for a CustomBlock with the same name on page 0 (hidden)
            if n in field_blocks: block = field_blocks[n]
            # otherwise, use the default text CustomBlock
            else: block = CustomBlock.text_create()
            fields.append((n, block.field()))
        
        name = self.program.db_slug + '_' + self.db_slug + '_i'
        class Meta:
            constraints = [
                UniqueConstraint(
                    fields=['_submission', '_collection', '_block', '_rank'],
                    name=self.program.db_slug+'_'+self.db_slug+'_u'
                )
            ]
            verbose_name = self.slug + ' item'
            verbose_name_plural = self.slug + ' items'
        return create_model(name, fields, program=self.program.db_slug,
                            base_class=SubmissionItem, meta=Meta)
    
    def publish_model(self, model, admin=None):
        with connection.schema_editor() as editor:
            editor.create_model(model)
        ctype = ContentType(app_label=model._meta.app_label,
                            model=model.__name__)
        ctype.save()
        ContentType.objects.clear_cache()
        
        from ..admin import site
        site.register(model, admin)
    
    def unpublish_model(self, model):
        from ..admin import site
        match = None
        for m in site._registry.keys():
            if m._meta.db_table == model._meta.db_table: match = m
        if match: site.unregister(match)
        
        ctype = ContentType.objects.get_for_model(model)
        ctype.delete()
        ContentType.objects.clear_cache()
        with connection.schema_editor() as editor:
            editor.delete_model(model)
            
    def publish(self):
        if self.status != self.Status.DRAFT: return
        
        self.status = self.Status.ENABLED
        if 'model' in self.__dict__: del self.model
        if 'item_model' in self.__dict__: del self.item_model
        
        from ..admin import SubmissionAdmin, SubmissionItemAdmin
        self.publish_model(self.model, admin=SubmissionAdmin)
        if self.item_model:
            self.publish_model(self.item_model, admin=SubmissionItemAdmin)
        
        self.modified = timezone.now()
        self.save()
    
    def unpublish(self):
        if self.status == self.Status.DRAFT: return
        
        self.unpublish_model(self.model)
        if self.item_model: self.unpublish_model(self.item_model)
        
        self.status = self.Status.DRAFT
        if 'model' in self.__dict__: del self.model
        if 'item_model' in self.__dict__: del self.item_model
        
        self.modified, self.completed = timezone.now(), None
        self.save()
    
    def get_available_plugins(self):
        from ..plugins import get_all_plugins
        
        return { plugin.module: plugin for plugin in get_all_plugins(self)
                 if not plugin.name.startswith('.') }
    
    def get_plugins(self):
        if 'plugins' in self.options: return self.options['plugins']
        return []
    
    def add_plugins(self, plugins):
        available = self.get_available_plugins()
        
        enable = [ p for p in plugins if p in available ]
        for plugin in enable:
            if hasattr(available[plugin].app, 'installed'):
                getattr(available[plugin].app, 'installed')(self)
        
        if 'plugins' not in self.options: self.options['plugins'] = []
        self.options['plugins'] += plugins
    
    def remove_plugins(self, plugins):
        available = self.get_available_plugins()

        for plugin in plugins:
            if hasattr(available[plugin].app, 'uninstalled'):
                getattr(available[plugin].app, 'uninstalled')(self)

        new_plugins = [ p for p in self.options['plugins'] if p not in plugins ]
        self.options['plugins'] = new_plugins
    
    def default_text_label_style(self):
        if 'default_text_label_style' in self.options:
            return self.options['default_text_label_style']
        return FormLabel.LabelStyle.WIDGET
    
    def num_pages(self):
        return self.blocks.aggregate(Max('page'))['page__max']
    
    def custom_blocks(self):
        return CustomBlock.objects.filter(form=self).non_polymorphic()
    
    def submission_blocks(self):
        blocks = self.blocks.not_instance_of(CollectionBlock)
        return blocks.exclude(page=0, _rank__gt=0)
    
    def collections(self, name=None):
        blocks = CollectionBlock.objects.filter(form=self)
        if name: return blocks.filter(name=name).non_polymorphic()
        return blocks.non_polymorphic()
    
    def validation_block(self):
        return self.blocks.get(page=0, _rank=1)
    
    def visible_blocks(self, page=None, skip=None):
        query = self.blocks.all()
        if skip: query = query.exclude(id__in=skip)
        if page and page > 0: return query.filter(page=page)
        else: return query.exclude(page=0, _rank__gt=0)
        return query.filter(page__gt=0)
    
    def visible_items(self, submission, page=None, skip=None):
        if not self.item_model: return []
        
        query = self.item_model.objects.filter(_submission=submission)
        if skip: query = query.exclude(_block__in=skip)
        if page and page > 0:
            block_ids = Subquery(self.blocks.filter(page=page).values('pk'))
            query = query.filter(_block__in=block_ids)
        
        query = query.exclude(_file='', _filesize__gt=0) # upload in progress
        return query.order_by('_collection', '_block', '_rank')

    def field_labels(self):
        labels = {}
        for label in self.labels.all():
            key, target = label.path, labels
            if '.' in label.path:
                base, key = label.path.split('.')
                if key[-1] == '_': base, key = base + '_', key[:-1]
                
                if base not in labels: labels[base] = {}
                target = labels[base]
            if key not in target: target[key] = {}
            target[key][label.style] = label

        return labels

    def label_class(self):
        return FormLabel
    
    def status_message(self):
        if self.status == self.Status.DRAFT:
            return 'NA'
        elif self.status == self.Status.DISABLED:
            return _('Not yet open for submissions')
        elif self.status == self.Status.COMPLETED:
            return _('Closed')
        return _('Open for submissions')
    
    def hidden(self):
        return self.status != self.Status.ENABLED or 'hidden' in self.options
    
    def access_enable(self):
        if 'access_enable' in self.options: return self.options['access_enable']
        return None
    
    def timed_completion(self):
        if 'timed_completion' in self.options:
            return self.options['timed_completion']
        return None
    
    def complete_submit_time(self):
        if 'complete_submit_time' in self.options:
            return self.options['complete_submit_time']
        return 5 # minutes
    
    def extra_time(self):
        extra = self.complete_submit_time()
        if timezone.now() - self.completed <= timedelta(minutes=extra):
            return True
        return False
    
    def markdown(self, text):
        md = self.program.markdown
        return md.reset().convert(text)
    
    def review_pre(self, prefix=''):
        name = prefix + 'review_pre'
        if name in self.options:
            md = self.program.markdown
            return mark_safe(self.markdown(self.options[name]))
        return ''
    
    def review_post(self):
        if 'review_post' in self.options:
            md = self.program.markdown
            return mark_safe(self.markdown(self.options['review_post']))
        return ''
    
    def submit_submission(self, submission):
        submission._submit()
        
        rec, created = SubmissionRecord.objects.get_or_create(
            program=self.program, form=self.slug, submission=submission._id,
            type=SubmissionRecord.RecordType.SUBMISSION
        )
        rec.text = submission._email
        rec.save()
        
        if self.item_model:
            dir = os.path.join(settings.MEDIA_ROOT, str(submission._id))
            if os.path.isdir(dir):
                Path(os.path.join(dir, 'submitted')).touch()
    
    def submitted_review_pre(self):
        return self.review_pre(prefix='submitted_')
    
    def review_after_submit(self):
        return 'no_review_after_submit' not in self.options
    
    def submit_button_label(self):
        if 'submit_button_label' in self.options:
            return self.options['submit_button_label']
        return 'submit'
    
    def thanks(self):
        if 'thanks' in self.options:
            md = self.program.markdown
            return mark_safe(self.markdown(self.options['thanks']))
    
    def emails(self):
        if 'emails' in self.options: return self.options['emails']
        return {}
    
    def email_names(self):
        names = list(self.emails())
        for name in ('confirmation', 'continue'):
            if name not in names: names.insert(0, name)
        return names
    
    def load_email_templates(self, n):
        subject = loader.get_template('formative/emails/' + n + '_subject.html')
        content = loader.get_template('formative/emails/' + n + '.html')
        return subject, content
    
    def email_templates(self):
        emails = self.emails()
        for name in ('continue', 'confirmation'):
            if name in emails: continue
            subject, content = self.load_email_templates(name)
            
            emails[name] = {'content': content.template.source,
                            'subject': subject.template.source.rstrip('\n')}
        return emails


class FormLabel(models.Model):
    class Meta:
        constraints = [
            UniqueConstraint(fields=['form', 'path', 'style'],
                             name='unique_path_style')
        ]
    
    class LabelStyle(models.TextChoices):
        WIDGET = 'widget', _('widget label')
        VERTICAL = 'vertical', _('vertical label')
        HORIZONTAL = 'horizontal', _('horizontal label')
    
    form = models.ForeignKey('Form', models.CASCADE, related_name='labels',
                             related_query_name='label')
    path = models.CharField(max_length=128)
    text = models.CharField(max_length=1000)
    style = models.CharField(max_length=16, choices=LabelStyle.choices,
                             default=LabelStyle.WIDGET)
    
    def __str__(self):
        return self.path
    
    def display(self, inline=False):
        s = self.form.program.markdown.reset().convert(self.text)
        if inline: return mark_safe(remove_p(s))
        return mark_safe(s)
    
    def display_inline(self):
        return self.display(inline=True)


class FormDependency(models.Model):
    class Meta:
        verbose_name = 'dependency value'
        verbose_name_plural = 'dependency values'
        constraints = [
            UniqueConstraint(fields=['block', 'value'], name='unique_blockval')
        ]
    
    block = models.ForeignKey('FormBlock', models.CASCADE,
                              related_name='dependencies',
                              related_query_name='dependency')
    value = models.CharField(max_length=64, blank=True)
    
    def __str__(self):
        if self.block.dependence:
            return f'{self.block.dependence.name}="{self.value}"'
        return f'?="{self.value}"'


class FormBlock(PolymorphicModel, RankedModel):
    class Meta(PolymorphicModel.Meta, RankedModel.Meta):
        constraints = [
            UniqueConstraint(fields=['form', 'page', '_rank'],
                             name='unique_rank'),
        ]
        ordering = ['form', 'page', '_rank']
    
    form = models.ForeignKey(Form, models.CASCADE,
                             related_name='blocks', related_query_name='block')
    name = models.SlugField(max_length=32, verbose_name='identifier',
                            allow_unicode=True)
    options = models.JSONField(default=dict, blank=True)
    page = models.PositiveIntegerField(default=1)
    dependence = models.ForeignKey('FormBlock', models.CASCADE,
                                   null=True, blank=True,
                                   related_name='dependents',
                                   related_query_name='dependent')
    negate_dependencies = models.BooleanField(default=False,
                                              verbose_name='negate dependency')
    
    def __str__(self):
        return self.name
    
    def rank_group(self):
        return FormBlock.objects.filter(form=self.form, page=self.page)
    
    def validate_unique(self, exclude=None):
        super().validate_unique(exclude=exclude)
        if not self._state.adding: return
        
        if self.name == 'email' and self.page:
            raise ValidationError('There is already a block called "email."')
        
        # name of a collection block identifies its "bucket", not its field(s)
        # in this case, it's not required to be unique
        if self.block_type() == 'collection': return
        
        qs = FormBlock.objects.filter(form_id=self.form_id)
        if self.page: qs = qs.filter(page__gt=0)
        else: qs = qs.filter(page=0)
        if qs.filter(name=self.name).exists():
            msg = 'Identifiers for stock and custom blocks must be unique.'
            raise ValidationError(msg)
    
    def block_type(self):
        if type(self) == CustomBlock: return 'custom'
        if type(self) == CollectionBlock: return 'collection'
        return 'stock'
    
    def stock_type(self):
        if 'type' not in self.options:
            raise FieldError('untyped stock widget')
        
        return StockWidget.by_type(self.options['type'])
    
    @cached_property
    def stock(self):
        return self.stock_type()(self.name, **self.options)
    
    def fields(self):
        return self.stock.fields()
    
    def enabled_blocks(self, value, page=None):
        # blocks on the given page that depend on self, and enabled given value
        query = self.form.blocks.filter(dependence_id=self.id)
        if page: query = query.filter(page=page)
        
        if type(value) == bool: value = value and 'yes' or 'no' # TODO: numeric
        if value is None: value = ''
        
        val = FormDependency.objects.filter(block_id=OuterRef('id'),
                                            value=Value(value))
        cond = Case(
            When(negate_dependencies=False, then=Exists(val)),
            When(negate_dependencies=True, then=~Exists(val))
        )
        query = query.annotate(en=cond).filter(en=True)
        return query.values_list('id', flat=True)
    
    def min_allowed_page(self):
        min_page = 1
        
        if self.dependence: min_page = self.dependence.page + 1
        return min_page
    
    def max_allowed_page(self, last_page=None):
        if last_page is None:
            last_page = self.form.blocks.aggregate(p=Max('page'))['p'] or 1
        max_page = last_page
        
        for block in self.dependents.all():
            if block.page - 1 < max_page: max_page = block.page - 1
        return max_page
    
    def show_in_review(self):
        return 'no_review' not in self.options


class CustomBlock(FormBlock):
    class Meta:
        db_table = 'formative_formcustomblock'
    
    CHOICE_VAL_MAXLEN = 64
    DEFAULT_TEXT_MAXLEN = 1000
    MAX_TEXT_MAXLEN = 65535
    
    class InputType(models.TextChoices):
        TEXT = 'text', _('text')
        NUMERIC = 'num', _('numeric')
        CHOICE = 'choice', _('multiple choice')
        BOOLEAN = 'bool', _('true/false choice')
    
    block = models.OneToOneField(FormBlock, on_delete=models.CASCADE,
                                 parent_link=True, primary_key=True)
    type = models.CharField(max_length=16, choices=InputType.choices,
                            default=InputType.TEXT)
    required = models.BooleanField(default=False)
    num_lines = models.PositiveIntegerField(default=1)
    min_chars = models.PositiveIntegerField(null=True, blank=True)
    max_chars = models.PositiveIntegerField(null=True, blank=True)
    min_words = models.PositiveIntegerField(null=True, blank=True)
    max_words = models.PositiveIntegerField(null=True, blank=True)
    
    @classmethod
    def text_create(cls, *args, **kwargs):
        if 'max_chars' not in kwargs:
            kwargs['max_chars'] = cls.DEFAULT_TEXT_MAXLEN
        return cls(*args, **kwargs, type=cls.InputType.TEXT)
    
    def choices(self):
        if 'choices' not in self.options or not self.options['choices']:
            raise FieldError('choices not defined')
        
        return self.options['choices']
    
    def field(self):
        # fields are NULL when we haven't yet reached the page, or if their
        # block had a dependency that wasn't met. non-required fields may
        # have a different way to record that no input was made, usually ''
        if self.type == self.InputType.TEXT:
            blank, max_chars = False, self.max_chars
            
            if not self.min_chars and not self.min_words: blank = True
            if not self.max_chars or self.max_chars > self.MAX_TEXT_MAXLEN:
                max_chars = self.MAX_TEXT_MAXLEN
            
            if self.num_lines > 1 or max_chars > self.DEFAULT_TEXT_MAXLEN:
                return models.TextField(null=True, max_length=max_chars,
                                        blank=blank)
            return models.CharField(null=True, max_length=self.max_chars,
                                    blank=blank)

        elif self.type == self.InputType.NUMERIC:
            return models.IntegerField(null=True, blank=(not self.required))

        elif self.type == self.InputType.CHOICE:
            return models.CharField(null=True, blank=(not self.required),
                                    max_length=self.CHOICE_VAL_MAXLEN,
                                    choices=[(c, c) for c in self.choices()])
        
        elif self.type == self.InputType.BOOLEAN:
            return models.BooleanField(null=True)
    
    def fields(self):
        return [(self.name, self.field())]

    def form_field(self, model_field, **kwargs):
        if self.type == self.InputType.TEXT:
            return model_field.formfield(min_length=self.min_chars, **kwargs)
        
        elif self.type == self.InputType.NUMERIC:
            return model_field.formfield(min_value=self.numeric_min(),
                                         max_value=self.numeric_max(), **kwargs)
        
        # or use the ModelForm factory's default:
        return model_field.formfield(**kwargs)
    
    def clean_field(self, data, field):
        # currently, all are handled from validators set up on the form
        return data
    
    def numeric_min(self):
        if 'numeric_min' in self.options: return self.options['numeric_min']
        return None
    
    def numeric_max(self):
        if 'numeric_max' in self.options: return self.options['numeric_max']
        return None
    
    def default_value(self):
        if self.type == self.InputType.TEXT: return None
        
        if 'default_value' in self.options: return self.options['default_value']
        return None
    
    def conditional_value(self, value):
        if self.type in (self.InputType.TEXT, self.InputType.NUMERIC):
            # in this case, condition is whether the field was filled out
            # (and is non-zero for numeric)
            return bool(value)
        
        return value
    
    def span(self, media=None):
        width = 6
        if self.max_chars and self.max_chars > 50: width = 8
        if self.num_lines > 1: width = 8
        if self.num_lines > 4: width = 10
        if self.type in (self.InputType.CHOICE, self.InputType.BOOLEAN):
            width = 8
        elif self.type == self.InputType.NUMERIC: width = 2
        
        if self.type == self.InputType.NUMERIC:
            if 'span_phone' in self.options:
                return min(self.options['span_phone'], 4)

        if 'span_tablet' in self.options:
            if not media: return min(width, self.options['span_tablet'], 4)
        elif not media: return min(width, 4)
        
        if media == 'tablet' and 'span_tablet' in self.options:
            return self.options['span_tablet']
        if media == 'desktop' and 'span_desktop' in self.options:
            return self.options['span_desktop']
        
        return width
    
    def tablet_span(self): return self.span(media='tablet')
    def desktop_span(self): return self.span(media='desktop')


class CollectionBlock(FormBlock):
    class Meta:
        db_table = 'formative_formcollectionblock'
    
    FIXED_CHOICE_VAL_MAXLEN = 100
    
    class AlignType(models.TextChoices):
        HORIZONTAL = 'horizontal', _('horizontal')
        VERTICAL = 'vertical', _('vertical')
    
    block = models.OneToOneField(FormBlock, on_delete=models.CASCADE,
                                 parent_link=True, primary_key=True)
    fixed = models.BooleanField(default=False)
    min_items = models.PositiveIntegerField(null=True, blank=True) # null if
    max_items = models.PositiveIntegerField(null=True, blank=True) # fixed
    has_file = models.BooleanField(default=False)
    file_optional = models.BooleanField(default=False)
    # we don't need these references indexed or validated here, so no SlugField
    name1 = models.CharField(max_length=32, default='', blank=True)
    name2 = models.CharField(max_length=32, default='', blank=True)
    name3 = models.CharField(max_length=32, default='', blank=True)
    align_type = models.CharField(max_length=16, choices=AlignType.choices,
                                  default=AlignType.HORIZONTAL)
    
    def fields(self):
        return []
    
    def collection_fields(self):
        fields = []
        if self.name1: fields.append(self.name1)
        if self.name2: fields.append(self.name2)
        if self.name3: fields.append(self.name3)
        
        return fields
    
    def max_filesize(self):
        if 'max_filesize' in self.options: return self.options['max_filesize']
        return None # TODO: overall default max
    
    def allowed_filetypes(self):
        if 'file_types' in self.options:
            return self.options['file_types']
        
        return None # allow any file extension
    
    def allowed_extensions(self):
        types = self.allowed_filetypes()
        if not types: return None
        
        extensions = []
        for filetype in types:
            extensions += FileType.by_type(filetype)().allowed_extensions()
        return extensions
    
    def autoinit_filename(self):
        if 'autoinit_filename' in self.options: return True
        return False
    
    def fixed_choices(self):
        if 'choices' not in self.options:
            msg = 'choices must be provided for a fixed collection block'
            raise FieldError(msg)
        
        return self.options['choices']
    
    def num_choices(self):
        return len(self.fixed_choices())
    
    def file_limits(self):
        if 'file_limits' in self.options: return self.options['file_limits']
        return {}
    
    def process_options(self, filetype):
        if 'file_processing' not in self.options: return {}
        if filetype in self.options['file_processing']:
            return self.options['file_processing'][filetype]
        return {}
    
    def span(self, media=None):
        width = 10
        if media == 'tablet': width = 8

        if not media: return 4
        
        if media == 'tablet' and 'span_tablet' in self.options:
            return max(4, self.options['span_tablet'])
        if media == 'desktop' and 'span_desktop' in self.options:
            return max(4, self.options['span_desktop'])
        
        return width
    
    def tablet_span(self): return self.span(media='tablet')
    def desktop_span(self): return self.span(media='desktop')
    
    def total_colspan(self):
         fields = self.collection_fields()
         if fields: return len(fields)
         return 1
    
    def horizontal_width(self, field):
        total = self.total_colspan()
        if 'wide' in self.options:
            total += len(self.options['wide'])
            if field in self.options['wide']: return 200.0 / total
        return 100.0 / total
    
    def collection_fields_as_blocks(self):
        class ColFieldBlock:
            def __init__(self, name, width):
                self.name, self.width = name, width
        
        fields = self.collection_fields()
        return [ ColFieldBlock(n, self.horizontal_width(n)) for n in fields ]
    
    def items_sortable(self):
        return 'unsortable' not in self.options
    
    def button_text(self):
        if 'button_text' in self.options: return self.options['button_text']
        
        if self.has_file and not self.file_optional:
            if self.max_items and self.max_items > 1: return _('add files')
            return _('add file')
        return _('add item')


class SubmissionRecord(models.Model):
    class Meta:
        constraints = [
            UniqueConstraint(fields=['submission', 'type'],
                             name='unique_submission_record_type')
        ]
    
    class RecordType(models.TextChoices):
        FILES = 'files', _('uploaded files')
        SUBMISSION = 'submission', _('form submission')
    
    program = models.ForeignKey(Program, models.SET_NULL, null=True, blank=True)
    form = models.SlugField(max_length=64, allow_unicode=True)
    submission = models.UUIDField(editable=False)
    type = models.CharField(max_length=32)
    recorded = models.DateTimeField(auto_now=True, verbose_name='recorded at')
    text = models.TextField(blank=True)
    number = models.PositiveBigIntegerField(null=True, blank=True)
    deleted = models.BooleanField(default=False)

# abstract classes, used as templates for the dynamic models:

class Submission(models.Model):
    class Meta:
        abstract = True
    
    _id = models.UUIDField(primary_key=True, default=uuid.uuid4,
                           editable=False)
    # valid up to page N:
    _valid = models.PositiveIntegerField(default=0, editable=False)
    # an array of N block id arrays, those skipped for form dependency not met:
    _skipped = models.JSONField(default=list, blank=True, editable=False)
    _created = models.DateTimeField(auto_now_add=True)
    _modified = models.DateTimeField(auto_now=True)
    _submitted = models.DateTimeField(null=True, blank=True)
    
    @classmethod
    def _get_form(cls):
        program_slug = cls._meta.program_slug
        slug = cls._meta.model_name[len(program_slug)+1:]
        return Form.objects.get(program__db_slug=program_slug, db_slug=slug)
    
    def __str__(self):
        if hasattr(self, '_email'): return self._email
        return str(self._id)
    
    def _get_absolute_url(self):
        form = self._get_form()
        args = {'program_slug': form.program.slug, 'form_slug': form.slug,
                'sid': self._id}
        return reverse('submission', kwargs=args)
    
    def _update_context(self, form, context):
        context['review_link'] = submission_link(self, form, rest='review')
        
        for block in form.visible_blocks():
            if block.block_type() == 'custom':
                context[block.name] = getattr(self, block.name)
            elif block.block_type() == 'stock':
                class Obj: pass
                obj = Obj()
                obj.__dict__ = { n: getattr(self, block.stock.field_name(n))
                                 for n in block.stock.widget_names() }
                if len(obj.__dict__) > 1: context[block.name] = obj
                else: context[block.name] = next(iter(obj.__dict__.values()))
    
    def _send_email(self, form, name, **kwargs):
        form_emails = form.emails()
        if name in form_emails:
            subject = Template(form_emails[name]['subject'])
            template = Template(form_emails[name]['content'])
        else: subject, template = form.load_email_templates(name)
        
        context = {
            'submission': self, 'form': form,
            'submission_link': submission_link(self, form)
        }
        if self._submitted: self._update_context(form, context)
        
        return send_email(template=template, to=self._email,
                          subject=subject, context=context, **kwargs)
    
    def _submit(self):
        self._submitted = timezone.now()
        self.save()
    
    def _collections(self, queryset=None, form=None):
        if not form: form = self._get_form()
        if not queryset: queryset = self._items.all()
        
        # form's order also orders blocks' items with the same collection name
        block = FormBlock.objects.filter(form=form, pk=OuterRef('_block'))
        queryset = queryset.annotate(page=Subquery(block.values('page')))
        queryset = queryset.annotate(block_rank=Subquery(block.values('_rank')))
        items = queryset.order_by('_collection', 'page', 'block_rank', '_rank')
        collections = groupby(items, key=lambda item: item._collection)
        return { k: list(items) for k, items in collections }


def file_path(instance, filename):
    return os.path.join(str(instance._submission_id), filename)

class SubmissionItem(UnderscoredRankedModel):
    class Meta:
        abstract = True
    
    _id = models.BigAutoField(primary_key=True, editable=False)
    # see Form.item_model() for _submission = models.ForeignKey(Submission)
    
    # the item's collection name == the name of the CollectionBlock
    _collection = models.CharField(max_length=32)
    
    # id of the collection block this item came from, as some may have same name
    _block = models.PositiveBigIntegerField()
    
    _file = models.FileField(upload_to=file_path, max_length=172, blank=True)
    _filesize = models.PositiveBigIntegerField(default=0)
    _filemeta = models.JSONField(default=dict, blank=True)
    _error = models.BooleanField(default=False)
    _message = models.CharField(max_length=100, default='', blank=True)
    
    @classmethod
    def _filename_maxlen(cls):
        # use 37 for directory uuid, 8 for possible alt name, 7 for order prefix
        return cls._meta.get_field('_file').max_length - 37 - 8 - 7
    
    @classmethod
    def _message_maxlen(cls):
        return cls._meta.get_field('_message').max_length
    
    def _rank_group(self):
        return self.__class__.objects.filter(_submission=self._submission,
                                             _collection=self._collection,
                                             _block=self._block)
    
    def _file_name(self):
        if not self._file: return None
        return self._file.name[self._file.name.index('/')+1:]
    
    def _file_type(self):
        if 'type' in self._filemeta: return self._filemeta['type']
        return ''
    
    def _artifact_url(self, name='thumbnail'):
        if not self._file: return None
        type = self._file_type()
        if not type: return None
        if name != 'thumbnail' or type not in ('image', 'video'):
            filetype = FileType.by_type(type)()
            return filetype.artifact_url(name, self._file.url)
        
        if type == 'image': return thumbnail_path(self._file.url)
        elif type == 'video': return thumbnail_path(self._file.url, ext='jpg')
        return None



import re
import types
from copy import copy
from django.template import Library, TemplateSyntaxError
from widget_tweaks.templatetags.widget_tweaks import silence_without_field, \
    ATTRIBUTE_RE, FieldAttributeNode

register = Library()


def _process_field_attributes(field, attr, process):
    params = re.split(r"(?<!:):(?!:)", attr, 1)
    attribute = params[0].replace("::", ":")
    value = params[1] if len(params) == 2 else True
    field = copy(field)

    if not hasattr(field, 'as_widget'):
        old_tag = field.tag

        def tag(self, wrap_label=False):
            attrs = self.data['attrs']
            process(self.parent_widget, attrs, attribute, value)
            html = old_tag(wrap_label=False)
            self.tag = old_tag
            return html

        field.tag = types.MethodType(tag, field)
        return field
    
    old_as_widget = field.as_widget

    def as_widget(self, widget=None, attrs=None, only_initial=False):
        attrs = attrs or {}
        process(widget or self.field.widget, attrs, attribute, value)
        if attribute == "type":
            self.field.widget.input_type = value
            del attrs["type"]
        html = old_as_widget(widget, attrs, only_initial)
        self.as_widget = old_as_widget
        return html

    field.as_widget = types.MethodType(as_widget, field)
    return field


@register.filter("attr")
@silence_without_field
def set_attr(field, attr):
    def process(widget, attrs, attribute, value):
        attrs[attribute] = value

    return _process_field_attributes(field, attr, process)


@register.filter("add_error_attr")
@silence_without_field
def add_error_attr(field, attr):
    if hasattr(field, "errors") and field.errors:
        return set_attr(field, attr)
    return field


@register.filter("append_attr")
@silence_without_field
def append_attr(field, attr):
    def process(widget, attrs, attribute, value):
        if attrs.get(attribute):
            attrs[attribute] += " " + value
        elif widget.attrs.get(attribute):
            attrs[attribute] = widget.attrs[attribute] + " " + value
        else:
            attrs[attribute] = value

    return _process_field_attributes(field, attr, process)


@register.filter("add_class")
@silence_without_field
def add_class(field, css_class):
    return append_attr(field, "class:" + css_class)


@register.tag
def render_field(parser, token):
    error_msg = (
        '%r tag requires a form field followed by a list of '
        'attributes and values in the form attr="value"'
        % token.split_contents()[0]
    )
    try:
        bits = token.split_contents()
        tag_name = bits[0]  # noqa
        form_field = bits[1]
        attr_list = bits[2:]
    except ValueError:
        raise TemplateSyntaxError(error_msg)

    form_field = parser.compile_filter(form_field)

    set_attrs = []
    append_attrs = []
    for pair in attr_list:
        match = ATTRIBUTE_RE.match(pair)
        if not match:
            raise TemplateSyntaxError(error_msg + ": %s" % pair)
        dct = match.groupdict()
        attr, sign, value = (
            dct["attr"],
            dct["sign"],
            parser.compile_filter(dct["value"]),
        )
        if sign == "=":
            set_attrs.append((attr, value))
        else:
            append_attrs.append((attr, value))

    return FieldAttributeNode(form_field, set_attrs, append_attrs)



from django.apps import AppConfig, apps
from django.core.exceptions import ImproperlyConfigured

import sys


def get_all_plugins(form=None):
    plugins = []
    for app in apps.get_app_configs():
        if hasattr(app, 'FormativePluginMeta'):
            meta = app.FormativePluginMeta
            meta.module, meta.app = app.name, app
            
            if hasattr(app, 'is_available') and form:
                if not app.is_available(form): continue
            
            plugins.append(meta)
    
    return plugins

def get_matching_plugin(module):
    for app in apps.get_app_configs():
        if hasattr(app, 'FormativePluginMeta'):
            if app.name == module or module.startswith(app.name + '.'):
                meta = app.FormativePluginMeta
                meta.module, meta.app = app.name, app
                
                return meta
    return None


class PluginConfig(AppConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        if not hasattr(self, 'FormativePluginMeta'):
            msg = 'Formative plugin config needs a FormativePluginMeta class.'
            raise ImproperlyConfigured(msg)
        
        if hasattr(self.FormativePluginMeta, 'compatibility'):
            import pkg_resources
            try:
                pkg_resources.require(self.FormativePluginMeta.compatibility)
            except pkg_resources.VersionConflict as e:
                print('Incompatible plugins found.')
                print(f'Plugin {self.name} requires {e.req}, '
                      f'but you installed {e.dist}.')
                sys.exit(1)
